#include "TMDlib.h"
// LHAPDF parton distribution function (C++ wrapper)
#include "LHAPDF/FortranWrappers.h"
#include "LHAPDF/LHAPDF.h"


#include <iostream>
#include <sstream>
#include <string>  
#include <fstream>
using namespace std;
/* Common Block Declarations */

int iset;
extern string pdfpath;

bool first_sbrs = true;

void TMDlib::sbrsPDF(int irep, int imode, double x, double kt, double p, double& uval, double& dval, double& sea, double& charm)
{
  //Declare variables here

  const int fMem=200;   //Maximum number of replicas
  const int fNX=100;    //Maximum number of x-grid points
  const int fNQ2=2;     //Maximum number of Q2-grid points
  const int fNkt=50;    //Maximum number of kt-grid points
  const int fNFL=4;     //Maximum number of TMD PDFs

  const int m=4;
  const int n=1;
  const int o=4;
  
  const double pi = 3.14159;

  const double xmindata=0.06;
  const double xmaxdata=0.41;
  const double ktmaxdata=1.0;

  static int nrep, nxgrid, nktgrid, nQ2grid;
  int idum;
  
  int krep;
  int minx, maxx, midx, ix;
  int minQ, maxQ, midQ, iQ2;
  int mink, maxk, midk, ik;

  int ix1a[m], ix2a[n], ix3a[o];
  
  static double xminim, xmaxim, ktminim, ktmaxim, Q2minim, Q2maxim;
  static double xgrid[fNX], ktgrid[fNkt], Q2grid[fNQ2];

  static double xTMD[fMem][fNX][fNkt][fNQ2][fNFL];
  static double Nptdv[fMem], Nptuv[fMem], Nptsea[fMem];
  static double alpha[fMem], sigma[fMem];
  static double Nktfv[fMem], Nktunfv[fMem], NktKs[fMem], NktKu[fMem];
  static double beta[fMem], Gamma[fMem], delta[fMem];
  static double dln[fMem], nPKm[fMem], nDKm[fMem];

  double xpdf[fNFL];
 
  double TMuval[fMem], TMdval[fMem], TMsea[fMem];

  double y, dy, x1, x2, x3;
  static double x1a[fNX], x2a[fNkt], x3a[fNQ2];
  double ya[m][n][o];

  double p2=pow(p,2);
  const double xcha=0.1;

  x1=x;
  x2=p2;
  x3=kt;

  string filename;
//  string pdfpath="../data/";

  //Define path to grid
  if (pdfpath!="") 
    {
      //cout << "TMDlib::sbrsPDF TMDlibPATH = " <<  pdfpath << endl;
    }
  else 
    { 
      cout << " NO TMDlibPATH  set " <<endl; 
      return ;
    }

  if (iset == 300005 ) 
    {
      if (imode==0||imode==2)
	{
	  filename=pdfpath+"SBRS-2013-TMDPDFs-par.dat";
	}

      else if (imode==1)
	{
	  filename=pdfpath+"SBRS-2013-TMDPDFs.dat";
	}
      else
	{
	  cout << "  Invalid value for imode " << endl;
	  cout << "  imode = 0: analytc form " << endl;
	  cout << "  imode = 1: interpolation on the grid " << endl;
	  cout << "  imode = 2: analytic form, bT space " << endl;
	}
    }
  else 
    {
      cout << "  TMDlib:sbrsPDF set not found; iset = " << iset << endl;
      return ;
    }
  
  //Read grid file
  if (first_sbrs) 
    {
      cout << "sbrsPDF: read grid from " << filename << endl;

      if(imode==0||imode==2)
	{
	  cout << "  Initialization of TMD PDF analytic form " << endl;

	  //Read best-fit parameters from file
	  ifstream myfile;
	  string textdum;
	  myfile.open(filename.c_str());
	  if(!myfile) {cout << " The file cannot be opened " << endl; exit(1);}
	  else
	    {
	      //Read header and print to screen
	      for (;;)
		{
		  getline(myfile,textdum);
		  if (textdum.find("MinMax:") != string::npos) break;
		  cout << textdum << endl;
		}
	      
	      //Read min and max values
	      myfile >> nrep >> textdum >> idum;
	      myfile >> xminim >> textdum >> xmaxim >> textdum >> ktminim >> textdum >> ktmaxim >> textdum >> Q2minim >> textdum >> Q2maxim;
	      
	      for (;;)
		{
		  getline(myfile,textdum);
		  if (textdum.find("Some") != string::npos) break;
		  cout << textdum << endl;
		}	     
	      myfile >> nrep;
	      for(int krepwrap=0; krepwrap<nrep; krepwrap++)
		{
		  myfile >> Nptdv[krepwrap] 
			 >> Nptuv[krepwrap] 
			 >> Nptsea[krepwrap]
			 >> alpha[krepwrap] >> sigma[krepwrap]
			 >> Nktfv[krepwrap] >> Nktunfv[krepwrap]
			 >> NktKs[krepwrap] >> NktKu[krepwrap]
			 >> beta[krepwrap] >> Gamma[krepwrap] >> delta[krepwrap]
			 >> dln[krepwrap] >> nPKm[krepwrap] >> nDKm[krepwrap];
		}
	    }
	  myfile.close();

	  //Initialize MSTW 68% CL fit
	  cout << "  Initialize collinear PDFs " << endl;
	  
	  const int SUBSET = 0;
	  const string NAME = "MSTW2008lo68cl";
	  LHAPDF::initPDFSet(NAME, LHAPDF::LHGRID, SUBSET);
	  
	}

      else if(imode==1)
	{
	  cout << "  Initialization of TMD PDF grid for interpolation " << endl;

	  ifstream myfile;
	  string textdum;
	  myfile.open(filename.c_str());
	  if(!myfile) {cout << " The file cannot be opened " << endl; exit(1);}
	  else
	    {
	      //Read header and print to screen
	      for (;;)
		{
		  getline(myfile,textdum);
		  if (textdum.find("MinMax:") != string::npos) break;
		  cout << textdum << endl;
		}
	      
	      //Read min and max values
	      myfile >> nrep >> textdum >> idum;
	      myfile >> xminim >> textdum >> xmaxim >> textdum >> ktminim >> textdum >> ktmaxim >> textdum >> Q2minim >> textdum >> Q2maxim;
	      
	      for (;;)
		{
		  getline(myfile,textdum);
		  if (textdum.find("Some") != string::npos) break;
		  cout << textdum << endl;
		}
	      
	      myfile >> nrep;
	      
	      myfile >> nxgrid;
	      
	      for(int i=0; i<nxgrid; i++)
		{
		  myfile >> xgrid[i];
		}
	      
	      myfile >> nktgrid;
	      
	      for(int i=0; i<nktgrid; i++)
		{
		  myfile >> ktgrid[i];
		}
	      
	      myfile >> nQ2grid;
	      
	      for(int i=0; i<nQ2grid; i++)
		{
		  myfile >> Q2grid[i];
		}
	      
	      myfile >> nrep;
	      
	      for(int ir=0; ir<nrep; ir++)
		{
		  for(int a=0; a<nxgrid; a++)
		    {
		      for (int b=0; b<nktgrid; b++)
			{
			  for(int c=0; c<nQ2grid; c++)
			    {
			      double PDF1, PDF2, PDF3, PDF4;
			      myfile >> xTMD[ir][a][b][c][0] 
				     >> xTMD[ir][a][b][c][1]
				     >> xTMD[ir][a][b][c][2]
				     >> xTMD[ir][a][b][c][3];
			    }
			}
		    }
		}
	      myfile >> textdum;
	    }
	  
	  myfile.close();
	}
      first_sbrs = false;
      cout << "  Parton densities read from file " << endl;
    }

  //Check interpolation range
  if (x<xminim || x>xmaxim)
    { 
      cout << "  Parton interpolation out of range: "
	   << " xmin = " << xminim << " xmax = " << xmaxim 
	   << " x = " << x << endl;
      exit(1);
    }
  if (kt<ktminim || kt > ktmaxim)
    {
      cout << "  Parton interpolation out of range: "
	   << " ktmin = " << ktminim << " GeV " 
	   << " ktmax = " << ktmaxim << " GeV " 
	   << " kt = " << kt << " GeV " << endl;
      exit(1);
    }
  if (p<sqrt(2.3)||p>sqrt(2.5))
    {
      cout << "  No evolution available: " 
	   << " Q2 = " << Q2minim << " GeV^2 " << " p = " << p<<  endl;
      exit(1);
    }

  //Check extrapolation region and print warnings
  if (x<xmindata || x>xmaxdata)
    {
      cout << "  Warning: extrapolation made outside data region" << endl;
      cout << "  The output may not be reliable at x = " << x << endl;
    }
  if (kt>ktmaxdata)
    {
      cout << "  Warning: extrapolation made outside data region" << endl;
      cout << "  The output may not be reliable at kt = " << kt << " GeV " << endl;
    }

  if(irep<0||irep>nrep)
    {
      cout << "invalid value for irep = " << irep << endl;
      exit(1);
    }

  //Return TMD-PDFs by using their analytic form
  if(imode==0)
    {

      //Compute average transverse momenta
      for(int krepwrap=0; krepwrap<nrep; krepwrap++)
	{	  
	  TMuval[krepwrap]= Nptuv[krepwrap]*( pow(1-x,alpha[krepwrap])*pow(x,sigma[krepwrap]) )/( pow(1-0.1,alpha[krepwrap])*pow(0.1,sigma[krepwrap]) );
	  TMdval[krepwrap] = Nptdv[krepwrap]*( pow(1-x,alpha[krepwrap])*pow(x,sigma[krepwrap]) )/( pow(1-0.1,alpha[krepwrap])*pow(0.1,sigma[krepwrap]) );
	  TMsea[krepwrap] = Nptsea[krepwrap]*( pow(1-x,alpha[krepwrap])*pow(x,sigma[krepwrap]) )/( pow(1-0.1,alpha[krepwrap])*pow(0.1,sigma[krepwrap]) );  
	}
      double Qmeanfit;
      Qmeanfit = p;

      uval = ( LHAPDF::xfx(x,Qmeanfit,2) - LHAPDF::xfx(x,Qmeanfit,-2) )*( (1/(pi*TMuval[irep]))*exp( -pow(kt,2)/TMuval[irep] ) );

      dval = ( LHAPDF::xfx(x,Qmeanfit,1) - LHAPDF::xfx(x,Qmeanfit,-1) )*( (1/(pi*TMdval[irep]))*exp( -pow(kt,2)/TMdval[irep] ) ); 

      sea = LHAPDF::xfx(x,Qmeanfit,3)*( (1/(pi*TMsea[irep]))*exp( -pow(kt,2)/TMsea[irep] ) );

      charm = LHAPDF::xfx(x,Qmeanfit,4)*( (1/(pi*TMsea[irep]))*exp( -pow(kt,2)/TMsea[irep] ) );

    }
  
  //Return TMD-PDFs by interpolating on the grid file
  else if(imode==1)
    {
      //Find nearest points for grid interpolation
      minx=0;
      maxx=nxgrid;
      do
	{
	  midx=(minx+maxx)/2;
	  if(x<xgrid[midx])
	    {maxx=midx;}
	  else
	    {minx=midx;}
	}
      while ((maxx-minx)>1);
      ix=minx;
      
      minQ=0;
      maxQ=nQ2grid;
      do
	{
	  midQ=(minQ+maxQ)/2;
	  if(p2<Q2grid[midQ])
	    {maxQ=midQ;}
	  else
	    {minQ=midQ;}
	}
      while ((maxQ-minQ)>1);
      iQ2=minQ;
      
      mink=0;
      maxk=nktgrid;
      do
	{
	  midk=(mink+maxk)/2;
	  if(kt<ktgrid[midk]) 
	    {maxk=midk;}
	  else
	    {mink=midk;}
	}
      while ((maxk-mink)>1);
      ik=mink; 
      
      for(int i=0; i<m; i++)
	{
	  if(ix>=m/2&&ix<=(nxgrid-m/2)) {ix1a[i]=ix-m/2+i;}
	  if(ix<m/2) {ix1a[i]=i;}
	  if(ix>(nxgrid-m/2)) {ix1a[i]=(nxgrid-m)+i;}
	  
	  if(ix1a[i]<0||ix1a[i]>=nxgrid)
	    {
	      cout << "Error in grids!" << endl;
	      cout << "i, ix1a[i] = " << i << "  " << ix1a[i] << endl;
	      exit(1);
	    }
	}

      for(int j=0; j<n; j++)
	{
	  if(iQ2>=n/2&&iQ2<=(nQ2grid-n/2)) {ix2a[j]=iQ2-n/2+j;}
	  if(iQ2<n/2) {ix2a[j]=j;}
	  if(iQ2>(nQ2grid-n/2)) {ix2a[j]=(nQ2grid-n)+j;}
	  
	  if(ix2a[j]<0||ix2a[j]>=nQ2grid)
	    {
	      cout << "Error in grids!" << endl;
	      cout << "j, ix2a[j] = " << j << "  " << ix2a[j] << endl;
	      exit(1);
	    }
	}
      
      for(int l=0; l<o; l++) 
	{
	  if(ik>=o/2&&ik<=(nktgrid-o/2)) {ix3a[l]=ik-o/2+l;}
	  if(ik<o/2) {ix3a[l]=l;}
	  if(ik>(nktgrid-o/2)) {ix3a[l]=(nktgrid-o)+l;}
	  
	  if(ix3a[l]<0||ix3a[l]>=nktgrid)
	    {
	      cout << "Error in grids!" << endl;
	      cout << "l, ix3a[l] = " << l << "  " << ix3a[l] << endl;
	      exit(1);
	    }
	}
            
      
      for(int jpdf=0; jpdf<fNFL; jpdf++)
	{
	  for(int i=0; i<m; i++)
	    {
	      x1a[i]=xgrid[ix1a[i]];

	      for(int j=0; j<n; j++)
		{
		  x2a[j]=Q2grid[ix2a[j]];

		  for(int l=0; l<o; l++)
		    {
		      x3a[l]=ktgrid[ix3a[l]];
		      ya[i][j][l]=xTMD[irep][ix1a[i]][ix3a[l]][ix2a[j]][jpdf];
		    }
		}
	    }
	  polin3(x1a,x2a,x3a,ya,m,n,o,x1,x2,x3,y,dy);
	  xpdf[jpdf]=y;
	}
      
      uval = xpdf[0];
      dval = xpdf[1];
      sea = xpdf[2];
      charm = xpdf[3];
    }

  //Return the FT in bT space for TMD-PDFs using the analytic result  
  if(imode==2)
    {
      
      //Compute average transverse momenta
      for(int krepwrap=0; krepwrap<nrep; krepwrap++)
	{	  
	  TMuval[krepwrap]= Nptuv[krepwrap]*( pow(1-x,alpha[krepwrap])*pow(x,sigma[krepwrap]) )/( pow(1-0.1,alpha[krepwrap])*pow(0.1,sigma[krepwrap]) );
	  TMdval[krepwrap] = Nptdv[krepwrap]*( pow(1-x,alpha[krepwrap])*pow(x,sigma[krepwrap]) )/( pow(1-0.1,alpha[krepwrap])*pow(0.1,sigma[krepwrap]) );
	  TMsea[krepwrap] = Nptsea[krepwrap]*( pow(1-x,alpha[krepwrap])*pow(x,sigma[krepwrap]) )/( pow(1-0.1,alpha[krepwrap])*pow(0.1,sigma[krepwrap]) );  
	}
      double Qmeanfit;
      Qmeanfit = sqrt(p);
      
      uval = ( LHAPDF::xfx(x,Qmeanfit,2) - LHAPDF::xfx(x,Qmeanfit,-2) )*exp(-pow(kt,2)*TMuval[irep]/4.0);

      dval = ( LHAPDF::xfx(x,Qmeanfit,1) - LHAPDF::xfx(x,Qmeanfit,-1) )*exp(-pow(kt,2)*TMdval[irep]/4.0 ); 

      sea = LHAPDF::xfx(x,Qmeanfit,3)*exp(-pow(kt,2)*TMsea[irep]/4.0 );

      charm = LHAPDF::xfx(x,Qmeanfit,4)*exp(-pow(kt,2)*TMsea[irep]/4.0 );

    }

  return ;

}
